#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 12
#endif
#ifndef PFXDIST
   #define PFXDIST 256
#endif
#ifndef PFADIST
   #define PFADIST 256
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGER2K(ATL_CINT M0, ATL_CINT N, const float *X, const float *Y, 
                const float *W, const float *Z, float *A, ATL_CINT lda0)
/*
 * 1x12 unrolled r2_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N12=(N/12)*12, M=M0+M0, M1=((M0/1)*1)<<1,
            lda=lda0+lda0, lda12=lda*12;
   ATL_INT j;

   for (j=N12; j; j -= 12, A += lda12, Y += 24)
   {
      float *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda, *A5=A4+lda,
            *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda, *A10=A9+lda,
            *A11=A10+lda;
      const register float ry0=Y[0], rz0=Z[0], iy0=Y[1], iz0=Z[1], ry1=Y[2],
                           rz1=Z[2], iy1=Y[3], iz1=Z[3], ry2=Y[4], rz2=Z[4],
                           iy2=Y[5], iz2=Z[5], ry3=Y[6], rz3=Z[6], iy3=Y[7],
                           iz3=Z[7], ry4=Y[8], rz4=Z[8], iy4=Y[9], iz4=Z[9],
                           ry5=Y[10], rz5=Z[10], iy5=Y[11], iz5=Z[11],
                           ry6=Y[12], rz6=Z[12], iy6=Y[13], iz6=Z[13],
                           ry7=Y[14], rz7=Z[14], iy7=Y[15], iz7=Z[15],
                           ry8=Y[16], rz8=Z[16], iy8=Y[17], iz8=Z[17],
                           ry9=Y[18], rz9=Z[18], iy9=Y[19], iz9=Z[19],
                           ry10=Y[20], rz10=Z[20], iy10=Y[21], iz10=Z[21],
                           ry11=Y[22], rz11=Z[22], iy11=Y[23], iz11=Z[23];
      ATL_INT i;
      prefY(Y+24+24-1);
      for (i=0; i < M1; i += 2)
      {
         register float ra00, ia00, ra10, ia10;
         const float rx0=X[i+0], rw0=W[i+0], ix0=X[i+1], iw0=W[i+1];

         prefX(X);

         ra00 = A0[i+0];
         ia00 = A0[i+0+1];
         ra00 += rx0 * ry0;
         ia00 += rx0 * iy0;
         ra00 -= ix0 * iy0;
         ia00 += ix0 * ry0;
         ra00 += rw0 * rz0;
         ia00 += rw0 * iz0;
         ra00 -= iw0 * iz0;
         A0[i+0] = ra00;
         ia00 += iw0 * rz0;
         A0[i+0+1] = ia00;


         ra00 = A1[i+0];
         ia00 = A1[i+0+1];
         ra00 += rx0 * ry1;
         ia00 += rx0 * iy1;
         ra00 -= ix0 * iy1;
         ia00 += ix0 * ry1;
         ra00 += rw0 * rz1;
         ia00 += rw0 * iz1;
         ra00 -= iw0 * iz1;
         A1[i+0] = ra00;
         ia00 += iw0 * rz1;
         A1[i+0+1] = ia00;


         ra00 = A2[i+0];
         ia00 = A2[i+0+1];
         ra00 += rx0 * ry2;
         ia00 += rx0 * iy2;
         ra00 -= ix0 * iy2;
         ia00 += ix0 * ry2;
         ra00 += rw0 * rz2;
         ia00 += rw0 * iz2;
         ra00 -= iw0 * iz2;
         A2[i+0] = ra00;
         ia00 += iw0 * rz2;
         A2[i+0+1] = ia00;


         ra00 = A3[i+0];
         ia00 = A3[i+0+1];
         ra00 += rx0 * ry3;
         ia00 += rx0 * iy3;
         ra00 -= ix0 * iy3;
         ia00 += ix0 * ry3;
         ra00 += rw0 * rz3;
         ia00 += rw0 * iz3;
         ra00 -= iw0 * iz3;
         A3[i+0] = ra00;
         ia00 += iw0 * rz3;
         A3[i+0+1] = ia00;


         ra00 = A4[i+0];
         ia00 = A4[i+0+1];
         ra00 += rx0 * ry4;
         ia00 += rx0 * iy4;
         ra00 -= ix0 * iy4;
         ia00 += ix0 * ry4;
         ra00 += rw0 * rz4;
         ia00 += rw0 * iz4;
         ra00 -= iw0 * iz4;
         A4[i+0] = ra00;
         ia00 += iw0 * rz4;
         A4[i+0+1] = ia00;


         ra00 = A5[i+0];
         ia00 = A5[i+0+1];
         ra00 += rx0 * ry5;
         ia00 += rx0 * iy5;
         ra00 -= ix0 * iy5;
         ia00 += ix0 * ry5;
         ra00 += rw0 * rz5;
         ia00 += rw0 * iz5;
         ra00 -= iw0 * iz5;
         A5[i+0] = ra00;
         ia00 += iw0 * rz5;
         A5[i+0+1] = ia00;


         ra00 = A6[i+0];
         ia00 = A6[i+0+1];
         ra00 += rx0 * ry6;
         ia00 += rx0 * iy6;
         ra00 -= ix0 * iy6;
         ia00 += ix0 * ry6;
         ra00 += rw0 * rz6;
         ia00 += rw0 * iz6;
         ra00 -= iw0 * iz6;
         A6[i+0] = ra00;
         ia00 += iw0 * rz6;
         A6[i+0+1] = ia00;


         ra00 = A7[i+0];
         ia00 = A7[i+0+1];
         ra00 += rx0 * ry7;
         ia00 += rx0 * iy7;
         ra00 -= ix0 * iy7;
         ia00 += ix0 * ry7;
         ra00 += rw0 * rz7;
         ia00 += rw0 * iz7;
         ra00 -= iw0 * iz7;
         A7[i+0] = ra00;
         ia00 += iw0 * rz7;
         A7[i+0+1] = ia00;


         ra00 = A8[i+0];
         ia00 = A8[i+0+1];
         ra00 += rx0 * ry8;
         ia00 += rx0 * iy8;
         ra00 -= ix0 * iy8;
         ia00 += ix0 * ry8;
         ra00 += rw0 * rz8;
         ia00 += rw0 * iz8;
         ra00 -= iw0 * iz8;
         A8[i+0] = ra00;
         ia00 += iw0 * rz8;
         A8[i+0+1] = ia00;


         ra00 = A9[i+0];
         ia00 = A9[i+0+1];
         ra00 += rx0 * ry9;
         ia00 += rx0 * iy9;
         ra00 -= ix0 * iy9;
         ia00 += ix0 * ry9;
         ra00 += rw0 * rz9;
         ia00 += rw0 * iz9;
         ra00 -= iw0 * iz9;
         A9[i+0] = ra00;
         ia00 += iw0 * rz9;
         A9[i+0+1] = ia00;


         ra00 = A10[i+0];
         ia00 = A10[i+0+1];
         ra00 += rx0 * ry10;
         ia00 += rx0 * iy10;
         ra00 -= ix0 * iy10;
         ia00 += ix0 * ry10;
         ra00 += rw0 * rz10;
         ia00 += rw0 * iz10;
         ra00 -= iw0 * iz10;
         A10[i+0] = ra00;
         ia00 += iw0 * rz10;
         A10[i+0+1] = ia00;


         ra00 = A11[i+0];
         ia00 = A11[i+0+1];
         ra00 += rx0 * ry11;
         ia00 += rx0 * iy11;
         ra00 -= ix0 * iy11;
         ia00 += ix0 * ry11;
         ra00 += rw0 * rz11;
         ia00 += rw0 * iz11;
         ra00 -= iw0 * iz11;
         A11[i+0] = ra00;
         ia00 += iw0 * rz11;
         A11[i+0+1] = ia00;

      }
      Z += 24;
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N12; j; j--, A += lda, Y += 2)
   {
      float *A0=A;
      const register float ry0=Y[0], rz0=Z[0], iy0=Y[1], iz0=Z[1];
      ATL_INT i;
      prefY(Y+24+24-1);
      for (i=0; i < M1; i += 2)
      {
         register float ra00, ia00, ra10, ia10;
         const float rx0=X[i+0], rw0=W[i+0], ix0=X[i+1], iw0=W[i+1];

         prefX(X);

         ra00 = A0[i+0];
         ia00 = A0[i+0+1];
         ra00 += rx0 * ry0;
         ia00 += rx0 * iy0;
         ra00 -= ix0 * iy0;
         ia00 += ix0 * ry0;
         ra00 += rw0 * rz0;
         ia00 += rw0 * iz0;
         ra00 -= iw0 * iz0;
         A0[i+0] = ra00;
         ia00 += iw0 * rz0;
         A0[i+0+1] = ia00;

      }
      Z += 2;
   }
}
